import * as Utils from 'conflux-web-utils';
import { randomHex, sha3, isHexStrict, hexToBytes } from 'conflux-web-utils';
import { formatters } from 'conflux-web-core-helpers';
import { ChainIdMethod, GetGasPriceMethod, GetTransactionCountMethod, AbstractMethodFactory } from 'conflux-web-core-method';
import isFunction from 'lodash/isFunction';
import isObject from 'lodash/isObject';
import Hash from 'cfx-lib/lib/hash';
import RLP from 'cfx-lib/lib/rlp';
import Bytes from 'cfx-lib/lib/bytes';
import { create, fromPrivate, sign, decodeSignature, encodeSignature, recover } from 'cfx-lib/lib/account';
import { AbstractConfluxWebModule } from 'conflux-web-core';
import scryptsy from 'scrypt.js';
import isString from 'lodash/isString';
import uuid from 'uuid';

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

var MethodFactory =
function (_AbstractMethodFactor) {
  _inherits(MethodFactory, _AbstractMethodFactor);
  function MethodFactory(utils, formatters) {
    var _this;
    _classCallCheck(this, MethodFactory);
    _this = _possibleConstructorReturn(this, _getPrototypeOf(MethodFactory).call(this, utils, formatters));
    _this.methods = {
      getChainId: ChainIdMethod,
      getGasPrice: GetGasPriceMethod,
      getTransactionCount: GetTransactionCountMethod
    };
    return _this;
  }
  return MethodFactory;
}(AbstractMethodFactory);

var crypto = typeof global === 'undefined' ? require('crypto-browserify') : require('crypto');
var Account =
function () {
  function Account(options) {
    var accounts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
    _classCallCheck(this, Account);
    this.address = options.address;
    this.privateKey = options.privateKey;
    this.accounts = accounts;
  }
  _createClass(Account, [{
    key: "signTransaction",
    value: function signTransaction(tx, callback) {
      return this.accounts.signTransaction(tx, this.privateKey, callback);
    }
  }, {
    key: "sign",
    value: function sign$1(data) {
      if (isHexStrict(data)) {
        data = hexToBytes(data);
      }
      var messageBuffer = Buffer.from(data);
      var preamble = "\x19Ethereum Signed Message:\n".concat(data.length);
      var preambleBuffer = Buffer.from(preamble);
      var cfxMessage = Buffer.concat([preambleBuffer, messageBuffer]);
      var hash = Hash.keccak256s(cfxMessage);
      var signature = sign(hash, this.privateKey);
      var vrs = decodeSignature(signature);
      return {
        message: data,
        messageHash: hash,
        v: vrs[0] === '0x' ? '0x00' : vrs[0],
        r: vrs[1],
        s: vrs[2],
        signature: signature
      };
    }
  }, {
    key: "encrypt",
    value: function encrypt(password, options) {
      return Account.fromPrivateKey(this.privateKey, this.accounts).toV3Keystore(password, options);
    }
  }, {
    key: "toV3Keystore",
    value: function toV3Keystore(password, options) {
      options = options || {};
      var salt = options.salt || crypto.randomBytes(32);
      var iv = options.iv || crypto.randomBytes(16);
      var derivedKey;
      var kdf = options.kdf || 'scrypt';
      var kdfparams = {
        dklen: options.dklen || 32,
        salt: salt.toString('hex')
      };
      if (kdf === 'pbkdf2') {
        kdfparams.c = options.c || 262144;
        kdfparams.prf = 'hmac-sha256';
        derivedKey = crypto.pbkdf2Sync(Buffer.from(password), salt, kdfparams.c, kdfparams.dklen, 'sha256');
      } else if (kdf === 'scrypt') {
        kdfparams.n = options.n || 8192;
        kdfparams.r = options.r || 8;
        kdfparams.p = options.p || 1;
        derivedKey = scryptsy(Buffer.from(password), salt, kdfparams.n, kdfparams.r, kdfparams.p, kdfparams.dklen);
      } else {
        throw new Error('Unsupported kdf');
      }
      var cipher = crypto.createCipheriv(options.cipher || 'aes-128-ctr', derivedKey.slice(0, 16), iv);
      if (!cipher) {
        throw new Error('Unsupported cipher');
      }
      var ciphertext = Buffer.concat([cipher.update(Buffer.from(this.privateKey.replace('0x', ''), 'hex')), cipher["final"]()]);
      var mac = sha3(Buffer.concat([derivedKey.slice(16, 32), Buffer.from(ciphertext, 'hex')])).replace('0x', '');
      return {
        version: 3,
        id: uuid.v4({
          random: options.uuid || crypto.randomBytes(16)
        }),
        address: this.address.toLowerCase().replace('0x', ''),
        crypto: {
          ciphertext: ciphertext.toString('hex'),
          cipherparams: {
            iv: iv.toString('hex')
          },
          cipher: options.cipher || 'aes-128-ctr',
          kdf: kdf,
          kdfparams: kdfparams,
          mac: mac.toString('hex')
        }
      };
    }
  }], [{
    key: "from",
    value: function from(entropy) {
      var accounts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      return new Account(create(entropy || randomHex(32)), accounts);
    }
  }, {
    key: "fromPrivateKey",
    value: function fromPrivateKey(privateKey) {
      var accounts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      return new Account(fromPrivate(privateKey), accounts);
    }
  }, {
    key: "fromV3Keystore",
    value: function fromV3Keystore(v3Keystore, password) {
      var nonStrict = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      var accounts = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
      if (!isString(password)) {
        throw new Error('No password given.');
      }
      var json = isObject(v3Keystore) ? v3Keystore : JSON.parse(nonStrict ? v3Keystore.toLowerCase() : v3Keystore);
      if (json.version !== 3) {
        throw new Error('Not a valid V3 wallet');
      }
      var derivedKey;
      var kdfparams;
      if (json.crypto.kdf === 'scrypt') {
        kdfparams = json.crypto.kdfparams;
        derivedKey = scryptsy(Buffer.from(password), Buffer.from(kdfparams.salt, 'hex'), kdfparams.n, kdfparams.r, kdfparams.p, kdfparams.dklen);
      } else if (json.crypto.kdf === 'pbkdf2') {
        kdfparams = json.crypto.kdfparams;
        if (kdfparams.prf !== 'hmac-sha256') {
          throw new Error('Unsupported parameters to PBKDF2');
        }
        derivedKey = crypto.pbkdf2Sync(Buffer.from(password), Buffer.from(kdfparams.salt, 'hex'), kdfparams.c, kdfparams.dklen, 'sha256');
      } else {
        throw new Error('Unsupported key derivation scheme');
      }
      var ciphertext = Buffer.from(json.crypto.ciphertext, 'hex');
      var mac = sha3(Buffer.concat([derivedKey.slice(16, 32), ciphertext])).replace('0x', '');
      if (mac !== json.crypto.mac) {
        throw new Error('Key derivation failed - possibly wrong password');
      }
      var decipher = crypto.createDecipheriv(json.crypto.cipher, derivedKey.slice(0, 16), Buffer.from(json.crypto.cipherparams.iv, 'hex'));
      var seed = "0x".concat(Buffer.concat([decipher.update(ciphertext), decipher["final"]()]).toString('hex'));
      return Account.fromPrivateKey(seed, accounts);
    }
  }]);
  return Account;
}();

var Wallet =
function () {
  function Wallet(utils, accountsModule) {
    _classCallCheck(this, Wallet);
    this.utils = utils;
    this.accountsModule = accountsModule;
    this.defaultKeyName = 'confluxweb_wallet';
    this.accounts = {};
    this.accountsIndex = 0;
    return new Proxy(this, {
      get: function get(target, name) {
        if (target.accounts[name]) {
          return target.accounts[name];
        }
        if (name === 'length') {
          return target.accountsIndex;
        }
        return target[name];
      }
    });
  }
  _createClass(Wallet, [{
    key: "create",
    value: function create(numberOfAccounts, entropy) {
      for (var i = 0; i < numberOfAccounts; ++i) {
        this.add(Account.from(entropy || this.utils.randomHex(32), this.accountsModule));
      }
      return this;
    }
  }, {
    key: "get",
    value: function get(account) {
      return this.accounts[account];
    }
  }, {
    key: "add",
    value: function add(account) {
      if (isString(account)) {
        account = Account.fromPrivateKey(account, this.accountsModule);
      }
      if (!this.accounts[account.address]) {
        account.index = this.accountsIndex;
        this.accounts[this.accountsIndex] = account;
        this.accounts[account.address] = account;
        this.accounts[account.address.toLowerCase()] = account;
        this.accountsIndex++;
        return account;
      }
      return this.accounts[account.address];
    }
  }, {
    key: "remove",
    value: function remove(addressOrIndex) {
      var account = this.accounts[addressOrIndex];
      if (account) {
        delete this.accounts[account.address];
        delete this.accounts[account.address.toLowerCase()];
        delete this.accounts[account.index];
        return true;
      }
      return false;
    }
  }, {
    key: "clear",
    value: function clear() {
      for (var i = 0; i <= this.accountsIndex; i++) {
        this.remove(i);
      }
      this.accountsIndex = 0;
      return this;
    }
  }, {
    key: "encrypt",
    value: function encrypt(password, options) {
      var encryptedAccounts = [];
      for (var i = 0; i < this.accountsIndex; i++) {
        encryptedAccounts.push(this.accounts[i].encrypt(password, options));
      }
      return encryptedAccounts;
    }
  }, {
    key: "decrypt",
    value: function decrypt(encryptedWallet, password) {
      var _this = this;
      encryptedWallet.forEach(function (keystore) {
        var account = Account.fromV3Keystore(keystore, password, false, _this.accountsModule);
        if (!account) {
          throw new Error("Couldn't decrypt accounts. Password wrong?");
        }
        _this.add(account);
      });
      return this;
    }
  }, {
    key: "save",
    value: function save(password, keyName) {
      console.warn('SECURITY WARNING: Storing of accounts in the localStorage is highly insecure!');
      if (typeof localStorage === 'undefined') {
        throw new TypeError('window.localStorage is undefined.');
      }
      try {
        localStorage.setItem(keyName || this.defaultKeyName, JSON.stringify(this.encrypt(password)));
      } catch (error) {
        if (error.code === 18) {
          return true;
        }
        throw new Error(error);
      }
      return true;
    }
  }, {
    key: "load",
    value: function load(password, keyName) {
      console.warn('SECURITY WARNING: Storing of accounts in the localStorage is highly insecure!');
      if (typeof localStorage === 'undefined') {
        throw new TypeError('window.localStorage is undefined.');
      }
      var keystore;
      try {
        keystore = localStorage.getItem(keyName || this.defaultKeyName);
        if (keystore) {
          keystore = JSON.parse(keystore);
        }
      } catch (error) {
        if (error.code === 18) {
          keystore = this.defaultKeyName;
        } else {
          throw new Error(error);
        }
      }
      return this.decrypt(keystore || [], password);
    }
  }]);
  return Wallet;
}();

var Accounts =
function (_AbstractConfluxWebMo) {
  _inherits(Accounts, _AbstractConfluxWebMo);
  function Accounts(provider, utils, formatters, methodFactory, options, net) {
    var _this;
    _classCallCheck(this, Accounts);
    _this = _possibleConstructorReturn(this, _getPrototypeOf(Accounts).call(this, provider, options, methodFactory, net));
    _this.utils = utils;
    _this.formatters = formatters;
    _this._transactionSigner = options.transactionSigner;
    _this.defaultKeyName = 'confluxweb_wallet';
    _this.accounts = {};
    _this.accountsIndex = 0;
    _this.wallet = new Wallet(utils, _assertThisInitialized(_this));
    return _this;
  }
  _createClass(Accounts, [{
    key: "create",
    value: function create(entropy) {
      return Account.from(entropy, this);
    }
  }, {
    key: "privateKeyToAccount",
    value: function privateKeyToAccount(privateKey) {
      return Account.fromPrivateKey(privateKey, this);
    }
  }, {
    key: "hashMessage",
    value: function hashMessage(data) {
      if (this.utils.isHexStrict(data)) {
        data = this.utils.hexToBytes(data);
      }
      var messageBuffer = Buffer.from(data);
      var preambleBuffer = Buffer.from("\x19Ethereum Signed Message:\n".concat(data.length));
      var cfxMessage = Buffer.concat([preambleBuffer, messageBuffer]);
      return Hash.keccak256s(cfxMessage);
    }
  }, {
    key: "signTransaction",
    value: function () {
      var _signTransaction = _asyncToGenerator(
      regeneratorRuntime.mark(function _callee(tx, privateKey, callback) {
        var account, signedTransaction;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.prev = 0;
                account = Account.fromPrivateKey(privateKey, this);
                if (tx.chainId) {
                  _context.next = 6;
                  break;
                }
                _context.next = 5;
                return this.getChainId();
              case 5:
                tx.chainId = _context.sent;
              case 6:
                if (tx.gasPrice) {
                  _context.next = 10;
                  break;
                }
                _context.next = 9;
                return this.getGasPrice();
              case 9:
                tx.gasPrice = _context.sent;
              case 10:
                if (tx.nonce) {
                  _context.next = 14;
                  break;
                }
                _context.next = 13;
                return this.getTransactionCount(account.address);
              case 13:
                tx.nonce = _context.sent;
              case 14:
                _context.next = 16;
                return this.transactionSigner.sign(this.formatters.inputCallFormatter(tx, this), account.privateKey);
              case 16:
                signedTransaction = _context.sent;
                if (isFunction(callback)) {
                  callback(false, signedTransaction);
                }
                return _context.abrupt("return", signedTransaction);
              case 21:
                _context.prev = 21;
                _context.t0 = _context["catch"](0);
                if (!isFunction(callback)) {
                  _context.next = 26;
                  break;
                }
                callback(_context.t0, null);
                return _context.abrupt("return");
              case 26:
                throw _context.t0;
              case 27:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this, [[0, 21]]);
      }));
      function signTransaction(_x, _x2, _x3) {
        return _signTransaction.apply(this, arguments);
      }
      return signTransaction;
    }()
  }, {
    key: "recoverTransaction",
    value: function recoverTransaction(rawTx) {
      var values = RLP.decode(rawTx);
      var signature = encodeSignature(values.slice(6, 9));
      var recovery = Bytes.toNumber(values[6]);
      var extraData = recovery < 35 ? [] : [Bytes.fromNumber(recovery - 35 >> 1), '0x', '0x'];
      var signingData = values.slice(0, 6).concat(extraData);
      var signingDataHex = RLP.encode(signingData);
      return recover(Hash.keccak256(signingDataHex), signature);
    }
  }, {
    key: "sign",
    value: function sign(data, privateKey) {
      if (this.utils.isHexStrict(data)) {
        data = this.utils.hexToBytes(data);
      }
      return Account.fromPrivateKey(privateKey, this).sign(data);
    }
  }, {
    key: "recover",
    value: function recover$1(message, signature, preFixed) {
      if (isObject(message)) {
        return this.recover(message.messageHash, encodeSignature([message.v, message.r, message.s]), true);
      }
      if (!preFixed) {
        message = this.hashMessage(message);
      }
      if (arguments.length >= 4) {
        return this.recover(arguments[0], encodeSignature([arguments[1], arguments[2], arguments[3]]), !!arguments[4]);
      }
      return recover(message, signature);
    }
  }, {
    key: "decrypt",
    value: function decrypt(v3Keystore, password, nonStrict) {
      return Account.fromV3Keystore(v3Keystore, password, nonStrict, this);
    }
  }, {
    key: "encrypt",
    value: function encrypt(privateKey, password, options) {
      return Account.fromPrivateKey(privateKey, this).toV3Keystore(password, options);
    }
  }, {
    key: "transactionSigner",
    get: function get() {
      return this._transactionSigner;
    }
    ,
    set: function set(transactionSigner) {
      if (transactionSigner.type && transactionSigner.type === 'TransactionSigner') {
        throw new Error('Invalid TransactionSigner given!');
      }
      this._transactionSigner = transactionSigner;
    }
  }]);
  return Accounts;
}(AbstractConfluxWebModule);

function Accounts$1(provider) {
  var net = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  return new Accounts(provider, Utils, formatters, new MethodFactory(Utils, formatters), options, net);
}

export { Accounts$1 as Accounts };
